use std::{path::PathBuf, sync::Arc};

use axum::{extract::{Path, State}, http::{response, StatusCode, Uri}, response::{Html, IntoResponse, Redirect}, routing::{get, post}, Form, Router};
use ldap3::{LdapConnAsync, LdapConnSettings};
use pandoc::{InputFormat, InputKind, MarkdownExtension, OutputKind};
use serde::{Deserialize, Serialize};
use sqlx::MySqlPool;
use strum::EnumIter;
use tera::{Context, Tera};

use tokio::sync::Mutex;
use tower_http::services::ServeDir; 

use tracing::{debug, error, Level};
use types::{form_applicant::FormApplicant, year_of_bachelor::YearOfBachelor};
use uuid::Uuid;
use validator::Validate;

pub mod types;

#[derive(Clone)]
struct AppState {
    tera: Tera,
    pool: MySqlPool,
    // conserve les données saisies pour les représenter au cas où elles ne sont pas validées
    form: Arc<Mutex<FormApplicant>>
}

impl AppState {
    async fn new() -> Self {
        let pool = sqlx::mysql::MySqlPool::connect(DB_URL).await.unwrap();

        let tera = match Tera::new("./templates/**/*.html") {
            Ok(t) => {
                for tpl in t.get_template_names() {
                    debug!("Added template: {tpl}");
                }
                debug!("Template loaded");
                t
            }
            Err(e) => {
                error!("Parsing errors: {e}");
                println!("Parsing errors: {e}");
                ::std::process::exit(1);
            }
        };

        AppState {
            pool,
            tera,
            form: Arc::new(Mutex::new(FormApplicant::new()))
        }
    }
}

const DB_URL: &str = "mysql://root:ntz-mdb2uch2nvc7DVW@enim-s-web2.enim.site.univ-lorraine.fr:3306/concours-v2";

#[tokio::main]
async fn main() {
    tracing_subscriber::fmt().with_max_level(Level::DEBUG).init();

    let state = AppState::new().await;

    // build our application with a single route
    let app = Router::new()
        .route("/", get(inscription))
        .route("/login", get(login))
        .route("/login", post(logged_in))
        .route("/inscription", get(inscription))
        .route("/add_applicant", post(add_applicant))
        .route("/recapitulation/:uuid", get(recapitulation))
        .fallback(fallback)
        // Permet de servir les fichiers PDF des dossiers générés (static file)
        .nest_service("/dossiers", ServeDir::new("./dossiers"))
        // .nest("dossiers", get(get_static_file))
        .with_state(state);

    // run our app with hyper, listening globally on port 3000
    let listener = tokio::net::TcpListener::bind("0.0.0.0:3000").await.unwrap();
    axum::serve(listener, app).await.unwrap();
}

async fn fallback(uri: Uri) -> (StatusCode, String) {
    (StatusCode::NOT_FOUND, format!("No route for {uri}"))
}

async fn inscription(State(state): State<AppState>) -> Html<String> {
    debug!("Inscription");
    debug!("FormCandidate: {:?}", state.form);
    let form = state.form.lock().await;
    let mut context = Context::from_serialize(form.clone()).unwrap();
    context.insert("year_of_bachelor", &YearOfBachelor::to_select_options());
    Html(state.tera.render(Template::Inscription.display(), &context).unwrap())
}

async fn login(State(state): State<AppState>) -> Html<String> {
    debug!("Login");
    Html(state.tera.render(Template::Login.display(), &Context::new()).unwrap())
}

#[derive(Serialize, Deserialize, Validate, Debug, Clone)]
struct Login {
    login: String,
    password: String,
}

#[derive(Clone,Serialize,Deserialize,Debug)]
enum LanguageSkill {
    Beginner,
    Intermediate,
    // Proficient,
    Fluent,
    // Native
}

impl ToString for LanguageSkill {
    fn to_string(&self) -> String {
        match self {
            LanguageSkill::Beginner => String::from("débutant"),
            LanguageSkill::Intermediate => String::from("intermédiaire"),
            LanguageSkill::Fluent => String::from("confirmé"),
        }
    }
}

#[derive(Serialize, Deserialize, Validate, Debug, Clone, sqlx::FromRow)]
struct Applicant {
    #[sqlx(rename="firstname")] 
    firstname: String,
    #[sqlx(rename="lastname")] 
    lastname: String,
    #[sqlx(rename="uuid")] 
    uuid: String,
}

#[derive(EnumIter)]
enum Template {
    Dossier,
    Inscription,
    Login,
    Pdf
}
impl Template {
    fn display(&self) -> &str {
        match self {
            Template::Pdf => "pdf.html",
            Template::Inscription => "inscription.html",
            Template::Login => "login.html",
            Template::Dossier => "dossier.html"
        }
    }
}

async fn logged_in(State(_state): State<AppState>, Form(_login): Form<Login>) -> Redirect {
    debug!("Logged in");
    let (conn, mut ldap) = LdapConnAsync::with_settings(LdapConnSettings::new()
        .set_starttls(false)
        .set_no_tls_verify(false),
        "ldaps://plg-dc2.ad.univ-lorraine.fr:636").await.unwrap();
    ldap3::drive!(conn);

    // Attempts a simple bind using the passed in values of username and Password
    let result = ldap.simple_bind("fery3@univ-lorraine.fr", "gutsberserk48").await.unwrap().success();
    if result.is_ok() {
        debug!("Auth OK");
    } else {
        debug!("Auth Failed");
    }
    // Ok(ldap.unbind());
    Redirect::to("/login")
}
async fn add_applicant(State(state): State<AppState>, Form(mut candidate): Form<FormApplicant>) -> Redirect {
    debug!("Add candidate");
    match candidate.validate() {
        Ok(_) => {
            let inserted_candidate = db_add_applicant(&state, candidate).await.unwrap();

            debug!("PDF candidate: {:?}", inserted_candidate);

            let rendered_data = state.tera.render(Template::Pdf.display(), &Context::from_serialize(&inserted_candidate).unwrap()).unwrap();

            let mut pandoc = pandoc::new();
            pandoc.set_input(InputKind::Pipe(rendered_data));
            pandoc.set_input_format(InputFormat::Html,[MarkdownExtension::RawHtml].to_vec());
            pandoc.set_output(OutputKind::File(PathBuf::from(format!("/home/fery3/opt/projectc/dossiers/{}.pdf", inserted_candidate.uuid))));
            pandoc.set_show_cmdline(true);
            let _ = pandoc.execute().unwrap();
            debug!("candidate validated");

            Redirect::to(&format!("/recapitulation/{}",inserted_candidate.uuid))
        },
        Err(e) => {
            // for error in e.field_errors() {
            //     let (field, _) = error;
            //     candidate.errors.push(field.to_string());
            // }
            candidate.errors = e.clone();
            debug!("add candidate error array {:?}", candidate.errors);

            let mut form = state.form.lock().await;
            *form = candidate.to_owned();
            // state.form = candidate;
            debug!("candidate not validated {e}");
            Redirect::to("/")
        }
    }
}

// async fn recapitulation(State(state): State<AppState>, Path(uuid): Path<String>) -> Html<String> {
async fn recapitulation(State(state): State<AppState>, Path(uuid): Path<String>) -> impl IntoResponse {

    let applicant = db_get_applicant(&state, uuid.clone());
    match applicant.await {
        Ok(applicant) => {
            Html(state.tera.render(Template::Dossier.display(), &Context::from_serialize(applicant).unwrap()).unwrap()).into_response()
        },
        Err(_) => {
            (StatusCode::NOT_FOUND, format!("No route for {uuid}")).into_response()
        }
    }
}

async fn db_add_applicant(state: &AppState, applicant: FormApplicant) -> Result<Applicant, sqlx::Error>{
    debug!("db_add_applicant: {:?}", applicant);
    let uuid = Uuid::new_v4();
    let insert = sqlx::query("insert into applicant
        (firstname, lastname, bea_number, scholarship_holder, email, mobile_phone, uuid)
        values (?, ?, ?, ?, ?, ?, ?)")
        .bind(&applicant.firstname)
        .bind(&applicant.lastname)
        .bind(&applicant.bea_number)
        .bind(applicant.scholarship_holder)
        .bind(&applicant.email)
        .bind(&applicant.mobile_phone)
        .bind(uuid.to_string())
        .execute(&state.pool).await;
    match insert {
        Ok(r) => {
            let userid = r.last_insert_id();
            sqlx::query_as::<_, Applicant>("select firstname, lastname, uuid from applicant where id = ?")
                .bind(userid)
                .fetch_one(&state.pool).await
        }
        Err(e) => {
            error!("add_applicant error: {}", e);
            Err(e)
        }
    }
}
async fn db_get_applicant(state: &AppState, token: String) -> Result<Applicant, sqlx::Error>{
    let applicant = sqlx::query_as::<_, Applicant>("select lastname, firstname, uuid from applicant where uuid = ?")
        .bind(token)
        .fetch_one(&state.pool).await;
    match applicant {
        Ok(c) => Ok(c),
        Err(e) => {
            error!("add_applicant error: {}", e);
            Err(e)
        }
    }
}

// #[cfg(test)]
// #[tokio::test]
// async fn test_sql_candidate() {
//     let state = AppState::new().await;
//     let candidate = FormCandidate {
//         lastname: "Fery".to_string(),
//         firstname: "Yann".to_string(),
//     };
//     let inserted_candidate = db_add_applicant(&state, candidate.clone()).await.unwrap();
//     assert_eq!(candidate, inserted_candidate);

// }

// #[tokio::test]
// async fn test_template_list() {
//     let state = AppState::new().await;

//     let mut tera: Vec<&str> = [].to_vec();
//     for tpl in state.tera.get_template_names() {
//         tera.push(tpl);
//     }
//     let mut list: Vec<String> = [].to_vec();
//     for tpl in Template::iter() {
//         list.push(tpl.display().to_string());
//     }
//     assert_eq!(tera, list);
// }
