use serde::{Deserialize, Serialize};
use validator::{Validate, ValidationError, ValidationErrors};

use super::academic_year::AcademicYear;

#[derive(Serialize, Deserialize, Validate, Debug, Clone, sqlx::FromRow)]
pub struct FormApplicant {
    #[validate(length(min=1, message="Ce champ est obligatoire"))]
    #[sqlx(rename="prenom")] 
    pub firstname: Option<String>,
    #[validate(length(min=1, message="Ce champ est obligatoire"))]
    #[sqlx(rename="nom")] 
    pub lastname: String,
    pub birthdate: String,
    pub birthplace: String,
    pub birthplace_dep: String,
    pub birthplace_country: String,
    pub citizenship: String,
    pub scholarship_holder: bool,
    #[validate(length(min=11, max=11, message="Le numéro doit faire exactement 11 caractères"))]
    pub bea_number: String, 
    #[validate(length(min=1))]
    pub address: String,
    #[validate(length(min=1))]
    pub postcode: String,
    #[validate(length(min=1))]
    pub city: String,
    pub country: String,
    #[validate(length(min=1, message="Numéro de téléphone obligatoire pour vous contacter"))]
    pub mobile_phone: String,
    #[validate(length(min=1, message="Adresse Email obligatoire pour vous contacter"))]
    pub email: String,
    pub year_n: Option<AcademicYear>,
    pub year_n_minus_1: Option<AcademicYear>,
    pub year_n_minus_2: Option<AcademicYear>,
    pub year_n_minus_3: Option<AcademicYear>,
    #[serde(skip_deserializing)]
    // pub errors: Vec<String>,
    pub errors: ValidationErrors,
}

impl FormApplicant {
    pub fn new() -> FormApplicant {
        FormApplicant {
            firstname: None,
            lastname: String::new(),
            birthdate: String::new(),
            birthplace: String::new(),
            birthplace_dep: String::new(),
            birthplace_country: String::new(),
            citizenship: String::new(),
            scholarship_holder: false,
            bea_number: String::new(), 
            // errors: Vec::new(),
            errors: ValidationErrors::new(),
            address: String::new(),
            postcode: String::new(),
            city: String::new(),
            country: String::new(),
            mobile_phone: String::new(),
            email: String::new(),
            year_n: None,
            year_n_minus_1: None,
            year_n_minus_2: None,
            year_n_minus_3: None,
        }
    }
}

impl Default for FormApplicant {
    fn default() -> Self {
        Self::new()
    }
}

impl PartialEq for FormApplicant {
    fn eq(&self, other: &Self) -> bool {
        self.firstname == other.firstname && self.lastname == other.lastname
    }
}
